# Copyright (c) 2012, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for ACPI"""

import acpi
import struct
import testsuite

def register_tests():
    testsuite.add_test("ACPI _MAT (Multiple APIC Table Entry) under Processor objects", test_mat, submenu="ACPI Tests")
    testsuite.add_test("ACPI _PSS (Pstate) tests", test_pss, submenu="ACPI Tests")

def test_mat():
    cpupaths = acpi.get_cpupaths()
    procid_apicid, uid_x2apicid = acpi.parse_madt()
    for cpupath in cpupaths:
        # Find the ProcId defined by the processor object
        processor = acpi.evaluate(cpupath)
        # Find the UID defined by the processor object's _UID method
        uid = acpi.evaluate(cpupath + "._UID")
        mat_buffer = acpi.evaluate(cpupath + "._MAT")
        if mat_buffer is None:
            continue
        # Process each _MAT subtable
        i = 0
        while i != len(mat_buffer):
            subtable = acpi.SubTable._make(struct.unpack_from(acpi.SubTable.fmt_str, mat_buffer, i))
            if subtable.Type == acpi.MADT_TYPE_LOCAL_APIC:
                localApic = acpi.MadtLocalApic._make(struct.unpack_from(acpi.MadtLocalApic.fmt_str, mat_buffer, i))
                if localApic.Flags & acpi.MADT_ENABLED:
                    if testsuite.test("{} Processor declaration ProcId = _MAT ProcId".format(cpupath), processor.ProcId == localApic.ProcId):
                        testsuite.print_detail("{} ProcId ({:#02x}) != _MAT ProcId ({:#02x})".format(cpupath, processor.ProcId, localApic.ProcId))
                        testsuite.print_detail("Processor Declaration: {}".format(processor))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, localApic))
                    if testsuite.test("{} with local APIC in _MAT has local APIC in MADT".format(cpupath), processor.ProcId in procid_apicid):
                        pass
                    elif testsuite.test("{} ApicId derived using Processor declaration ProcId = _MAT ApicId".format(cpupath), procid_apicid[processor.ProcId] == localApic.ApicId):
                        testsuite.print_detail("{} ApicId derived from MADT ({:#02x}) != _MAT ApicId ({:#02x})".format(cpupath, procid_apicid[processor.ProcId], localApic.ApicId))
                        testsuite.print_detail("Processor Declaration: {}".format(processor))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, localApic))
            if subtable.Type == acpi.MADT_TYPE_LOCAL_X2APIC:
                x2Apic = acpi.MadtLocalx2Apic._make(struct.unpack_from(acpi.MadtLocalx2Apic.fmt_str, mat_buffer, i))
                if x2Apic.Flags & acpi.MADT_ENABLED:
                    if testsuite.test("{} with x2Apic in _MAT has _UID".format(cpupath), uid is not None):
                        pass
                    elif testsuite.test("{}._UID = _MAT UID".format(cpupath), uid == x2Apic.UID):
                        testsuite.print_detail("{}._UID ({:#x}) != _MAT UID ({:#x})".format(cpupath, uid, x2Apic.UID))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, x2Apic))
                    if testsuite.test("{} with _MAT x2Apic has x2Apic in MADT".format(cpupath), x2Apic.UID in uid_x2apicid):
                        pass
                    elif testsuite.test("{} x2ApicId derived from MADT using UID = _MAT x2ApicId".format(cpupath), uid_x2apicid[x2Apic.UID] == x2Apic.X2ApicId):
                        testsuite.print_detail("{} x2ApicId derived from MADT ({:#02x}) != _MAT x2ApicId ({:#02x})".format(cpupath, uid_x2apicid[x2Apic.UID], x2Apic.X2ApicId))
                        testsuite.print_detail("_MAT offset[{}]: {}".format(i, x2Apic))
            i += subtable.Length

def test_pss():
    uniques = acpi.parse_pss()
    for pstates, cpupaths in uniques.iteritems():
        if testsuite.test("_PSS must exist", pstates is not None):
            if pstates is None:
                testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
                testsuite.print_detail('No _PSS exists')
                continue
        if testsuite.test("_PSS must not be empty", pstates != ()):
            if pstates == ():
                testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
                testsuite.print_detail('_PSS is empty')
                continue
        if testsuite.test("_PSS must contain at most 16 Pstates", len(pstates) <= 16):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            map(testsuite.print_detail, acpi.show_pss(pstates))
        frequencies = [p.CoreFrequency for p in pstates]
        if testsuite.test("_PSS must list Pstates in descending order of frequency", frequencies == sorted(frequencies, reverse=True)):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            map(testsuite.print_detail, acpi.show_pss(pstates))

def test_psd_thread_scope():
    uniques = acpi.parse_psd()
    if testsuite.test("_PSD (P-State Dependency) must exist for each processor", None not in uniques):
        if None not in uniques:
            testsuite.print_detail('_PSD exists for each processor')
        else:
            testsuite.print_detail(acpi.factor_commonprefix(uniques[None]))
            testsuite.print_detail('No _PSD exists')
            return
    unique_NumEntries = {}
    unique_Revision = {}
    unique_Domain = {}
    unique_CoordType = {}
    unique_NumProcessors = {}
    for value, cpupaths in uniques.iteritems():
        unique_NumEntries.setdefault(value.NumEntries, []).extend(cpupaths)
        unique_Revision.setdefault(value.Revision, []).extend(cpupaths)
        unique_Domain.setdefault(value.Domain, []).extend(cpupaths)
        unique_CoordType.setdefault(value.CoordType, []).extend(cpupaths)
        unique_NumProcessors.setdefault(value.NumProcessors, []).extend(cpupaths)

    def detail(d, fmt):
        for value, cpupaths in sorted(d.iteritems(), key=(lambda (k,v): v)):
            testsuite.print_detail(acpi.factor_commonprefix(cpupaths))
            testsuite.print_detail(fmt.format(value))

    if testsuite.test('_PSD.NumEntries must be 5', unique_NumEntries.keys() == [5]):
        detail(unique_NumEntries, 'Numtries = {} (Expected 5)')
    if testsuite.test('_PSD.Revision must be 0', unique_Revision.keys() == [0]):
        detail(unique_Revision, 'Revision = {}')
    if testsuite.test('_PSD.CoordType must be 0xFE (HW_ALL)', unique_CoordType.keys() == [0xfe]):
        detail(unique_CoordType, 'CoordType = {:#x} (Expected 0xFE HW_ALL)')
    if testsuite.test('_PSD.Domain must be unique (thread-scoped) for each processor', len(unique_Domain) == len(acpi.get_cpupaths())):
        detail(unique_Domain, 'Domain = {:#x} (Expected a unique value for each processor)')
    if testsuite.test('_PSD.NumProcessors must be 1', unique_NumProcessors.keys() == [1]):
        detail(unique_NumProcessors, 'NumProcessors = {} (Expected 1)')
